"""Utility functions related to calculating and constructing Fourier filters."""

from typing import Optional

import torch


def _handle_dim(dim: int | tuple[int, ...], ndim: int) -> tuple[int, ...]:
    """Convert dim to positive and tuple.

    Parameters
    ----------
    dim: int | tuple[int, ...]
        Tuple or single int of dimensions. Negative dimensions are indexed from the end.
    ndim: int
        Number of dimensions in the input tensor.

    Returns
    -------
    tuple[int, ...]
        Positive tuple of dimensions.

    Raises
    ------
    ValueError
        If any dimension is out of bounds.
    """
    if isinstance(dim, int):
        dim = (dim,)
    dim = tuple(d if d >= 0 else ndim + d for d in dim)

    if any(d < 0 or d >= ndim for d in dim):
        raise ValueError(f"Invalid dimensions {dim} for tensor of rank {ndim}")

    return dim


def _interp1d(
    x: torch.Tensor,
    xp: torch.Tensor,
    fp: torch.Tensor,
    left: Optional[float | complex] = None,
    right: Optional[float | complex] = None,
) -> torch.Tensor:
    """Helper func for unbatched 1D linear interpolation in torch matching np.interp.

    Parameters
    ----------
    x : torch.Tensor
        The x-coordinates at which to evaluate the interpolated values.
    xp : torch.Tensor
        The x-coordinates of the data points.
    fp : torch.Tensor
        The y-coordinates of the data points, must be same length as xp.
    left : float | complex, optional
        Value to return for x < xp[0], default is fp[0].
    right : float | complex, optional
        Value to return for x > xp[-1], default is fp[-1].

    Returns
    -------
    torch.Tensor
        The interpolated values, same shape as x.
    """
    # left and right fill values
    if left is None:
        left = fp[0]
    if right is None:
        right = fp[-1]

    # find the indices of the left and right values
    i = torch.searchsorted(xp, x, side="right")

    # Out of bounds masks
    oob_left = x < xp[0]
    oob_right = x >= xp[-1]

    # index-in-bounds (iib) and x-in-bounds (xib)
    iib = i[(~oob_left) & (~oob_right)]
    xib = x[(~oob_left) & (~oob_right)]

    f_lo = fp[iib - 1]
    f_hi = fp[iib]
    x_lo = xp[iib - 1]
    x_hi = xp[iib]

    # Construct output tensor and do liner interpolation
    out = torch.empty_like(x)
    out[(~oob_left) & (~oob_right)] = f_lo + (f_hi - f_lo) * (xib - x_lo) / (
        x_hi - x_lo
    )
    out[oob_left] = left
    out[oob_right] = right

    return out


def torch_interp(
    x: torch.Tensor,
    xp: torch.Tensor,
    fp: torch.Tensor,
    left: Optional[float | complex] = None,
    right: Optional[float | complex] = None,
    period: Optional[float] = None,
) -> torch.Tensor:
    """Reimplementation of np.interp in PyTorch for 1d linear interpolation.

    This function supports batching where the last dimensions of xp and fp are treated
    as the 1-dimensional data points and all other dimensions are batch dimensions.
    The x-coordinates to evaluate the interpolated values at (x) must still be
    1-dimensional. NOTE: Batch evaluation of interpolation is currently implemented as
    a naive for loop over dimensions. This may be updated in the future.

    NOTE: Period portion of function is not implemented.

    Parameters
    ----------
    x : torch.Tensor
        The x-coordinates at which to evaluate the interpolated values.
    xp : torch.Tensor
        The x-coordinates of the data points.
    fp : torch.Tensor
        The y-coordinates of the data points, must be same length as xp.
    left : float | complex, optional
        Value to return for x < xp[0], default is fp[0].
    right : float | complex, optional
        Value to return for x > xp[-1], default is fp[-1].
    period : float, optional
        Not implemented!

    Returns
    -------
    torch.Tensor
        The interpolated values, same shape as x.

    Raises
    ------
    NotImplementedError
        If period is not None.
    """
    if period is not None:
        raise NotImplementedError("Periodic interpolation is not implemented.")

    if x.ndim != 1:
        raise ValueError("Sample points (x) be 1D tensor.")
    if xp.shape != fp.shape:
        raise ValueError("Coordinates (xp) and values (fp) must have the same shape.")

    _is_batched = True
    if xp.ndim == 1:
        _is_batched = False
        xp = xp.unsqueeze(0)
        fp = fp.unsqueeze(0)

    # Naive for loop over batch dimensions
    n_batch = torch.prod(torch.tensor(xp.shape[:-1])).item()
    out = torch.zeros((n_batch, x.shape[0]), dtype=fp.dtype, device=fp.device)
    for i, (_xp, _fp) in enumerate(zip(xp.view(n_batch, -1), fp.view(n_batch, -1))):
        out[i] = _interp1d(x, _xp, _fp, left=left, right=right)

    # Squeeze batch dimension if input was not batched
    if not _is_batched:
        out = out.squeeze(0)

    return out


def curve_1dim_to_ndim(
    values_1d: torch.Tensor,  # shape (batch, N) or (N,)
    frequency_1d: torch.Tensor,  # shape (N,)
    frequency_grid: torch.Tensor,
    fill_lower: Optional[float | complex] = None,
    fill_upper: Optional[float | complex] = None,
) -> torch.Tensor:
    """Use interpolation to map 1D values from a curve to a nD grid.

    Parameters
    ----------
    values_1d : torch.Tensor
        Values of the one-dimensional curve corresponding to each frequency value.
    frequency_1d : torch.Tensor
        Corresponding frequency values of the one-dimensional curve.
    frequency_grid : torch.Tensor
        Absolute frequency values of the grid.
    fill_lower : float | complex, optional
        Fill value for lower (left) out-of-bound frequencies. Default is None and
        corresponds to the lowest (leftmost) value in values_1d.
    fill_upper : float | complex, optional
        Fill value for upper (right) out-of-bound frequencies. Default is None and
        corresponds to the highest (rightmost) value in values_1d.

    Returns
    -------
    torch.Tensor
        nD filter constructed from the 1D values.

    Example
    -------
    >>> values_1d = torch.tensor([1.0, 2.0, 7.0])
    >>> frequency_1d = torch.tensor([0.0, 0.25, 0.5])
    >>> grid = torch.tensor([[0.0. 0.1, 0.2], [0.3, 0.4, 0.5]])  # 2D toy example
    >>> filter_2d = curve_1dim_to_ndim(values_1d, frequency_1d, grid)
    >>> filter_2d
    tensor([[1.0, 1.5, 2.0],
            [4.0, 5.5, 7.0]])
    """
    if values_1d.ndim > 2:
        raise ValueError("values_1d must be at most 2D tensor.")

    _is_batched = True
    if values_1d.ndim == 1:
        values_1d = values_1d.unsqueeze(0)
        _is_batched = False

    interpn_values = torch.empty(values_1d.shape[0], *frequency_grid.shape)

    for i in range(values_1d.shape[0]):
        interpn_values[i] = torch_interp(
            frequency_grid.ravel(),
            frequency_1d,
            values_1d[i],
            left=fill_lower,
            right=fill_upper,
        ).reshape(frequency_grid.shape)

    # Squeeze batch dimension if input was not batched
    if not _is_batched:
        interpn_values = interpn_values.squeeze(0)

    return interpn_values


def bin_1dim_with_lerp(
    x: torch.Tensor,
    y: torch.Tensor,
    xp: torch.Tensor,
    normalize_by_count: bool = True,
) -> torch.Tensor:
    """Bins values (y) at coordinates (x) onto 1D grid (xp) with linear interpolation.

    NOTE: that multi-dimensional inputs (x, y) are supported, but batch dimensions
    are currently absent. All dimensions of the input will be flattened before binning.

    Parameters
    ----------
    x : torch.Tensor
        Coordinates of input values.
    y : torch.Tensor
        Values to bin.
    xp : torch.Tensor
        Coordinates of the 1D grid to bin onto. Note that xp *must* be a strictly
        increasing sequence *and* part of this function assumes that it has a linear
        spacing.
    normalize_by_count : bool, optional
        If True, the output values will be normalized by the number of partial values
        that fall in each bin. Default is True. If False, then the raw histogram-like
        values will be returned.

    Returns
    -------
    torch.Tensor
        Binned values on the 1D grid (histogram-like)
    """
    if x.shape != y.shape:
        raise ValueError("x and y must have the same shape.")

    if xp.ndim != 1:
        raise ValueError("xp must be 1D.")

    x = x.view(-1)
    y = y.view(-1)

    # Out of bounds masks to prevent indexing errors
    oob_left = x < xp[0]
    oob_right = x >= xp[-1] + xp[1] - xp[0]  # Assumes linear spacing
    x = x[~(oob_left | oob_right)]
    y = y[~(oob_left | oob_right)]

    # Find the nearest indices for binning values
    indices = torch.bucketize(input=x, boundaries=xp, right=False)

    # Find which 1d grid values each bin corresponds to
    left_bins = torch.clamp(indices - 1, 0, len(xp) - 1)
    right_bins = torch.clamp(indices, 0, len(xp) - 1)
    left_xp = xp[left_bins]
    right_xp = xp[right_bins]

    right_weights = (x - left_xp) / (right_xp - left_xp).clamp(min=1e-6)

    # Create sum and count tensors
    values_sum = torch.zeros_like(xp)
    values_count = torch.zeros_like(xp)

    # Accumulate values and counts
    values_sum.index_add_(0, left_bins, y * (1 - right_weights))
    values_sum.index_add_(0, right_bins, y * right_weights)

    if not normalize_by_count:
        return values_sum

    # Count number of partial values in each bin
    values_count.index_add_(0, left_bins, 1 - right_weights)
    values_count.index_add_(0, right_bins, right_weights)

    # Prevent division by zero
    values_count[values_count == 0] = 1

    return values_sum / values_count



"""MTF functions for the Fourier filter."""

import starfile
import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def read_mtf(file_path: str) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Read the MTF from a star file.

    Args:
        file_path: The path to the star file.

    Returns
    -------
        The frequencies (0 to 0.5 at Nyquist) and MTF amps (MTF(0)=1).
    """
    df = starfile.read(file_path)
    frequencies = torch.tensor(df["rlnResolutionInversePixel"].to_numpy()).float()
    mtf_amplitudes = torch.tensor(df["rlnMtfValue"].to_numpy()).float()

    return frequencies, mtf_amplitudes


def make_mtf_grid(
    image_shape: tuple[int, int] | tuple[int, int, int],
    mtf_frequencies: torch.Tensor,  # 1D tensor
    mtf_amplitudes: torch.Tensor,  # 1D tensor
    rfft: bool = True,
    fftshift: bool = False,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Make a MTF grid from the frequencies and amplitudes.

    Args:
        image_shape: The shape of the image.
        pixel_size: The pixel size of the image.
        mtf_frequencies: The frequencies of the MTF.
        mtf_amplitudes: The amplitudes of the MTF.
        rfft: If the FFT is a real FFT.
        fftshift: If the FFT is shifted.
        device: The device to use for the calculation.

    Returns
    -------
        The MTF grid.
    """
    # Calculate the FFT frequency grid (normalized to physical units)
    fft_freqgrid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=True,
        device=device,
    )  # Shape: (*image_shape), this return the frequency grid in range 0-0.5

    # Now map the MTF ampltiudes onto the corresponding frequencies
    # in the 2/3D fft_freq_px using linear interpolation

    # Clamp to the range of mtf_frequencies to avoid extrapolation issues
    fft_freq_mag_clamped = torch.clamp(
        fft_freqgrid, min=mtf_frequencies.min(), max=mtf_frequencies.max()
    )

    # Perform linear interpolation
    # Using torch.searchsorted to find bins
    indices = torch.searchsorted(mtf_frequencies, fft_freq_mag_clamped.flatten())
    indices = torch.clamp(indices, 1, len(mtf_frequencies) - 1)

    # Compute weights for linear interpolation
    x0 = mtf_frequencies[indices - 1]
    x1 = mtf_frequencies[indices]
    y0 = mtf_amplitudes[indices - 1]
    y1 = mtf_amplitudes[indices]
    weights = (fft_freq_mag_clamped.flatten() - x0) / (x1 - x0)

    # Interpolate amplitudes
    interpolated_amplitudes = y0 + weights * (y1 - y0)

    # Reshape to the original frequency grid shape
    mtf = interpolated_amplitudes.view(*fft_freq_mag_clamped.shape)

    return mtf



"""Envelope functions for the Fourier filter."""

import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid

from torch_fourier_filter.ctf import calculate_relativistic_electron_wavelength


def b_envelope(
    B: float,
    image_shape: tuple[int, int] | tuple[int, int, int],
    pixel_size: float,
    rfft: bool = True,
    fftshift: bool = False,
    device: torch.device = None,
) -> torch.Tensor:
    """
    Create a B-factor envelope for a Fourier transform.

    Parameters
    ----------
    B: float
        The B-factor value.
    image_shape: tuple[int, ...]
        Shape of the real space the dft is from input image.
    pixel_size: float
        The pixel size of the image.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False).
    fftshift: bool
        Whether the input is fftshifted.
    device: torch.device
        Device to place tensors on.

    Returns
    -------
    torch.Tensor
        B-factor envelope
    """
    frequency_grid_px = (
        fftfreq_grid(
            image_shape=image_shape,
            rfft=rfft,
            fftshift=fftshift,
            norm=True,
            device=device,
        )
        / pixel_size
    )

    divisor = 4  # this is 4 for amplitude, 2 for intensity
    b_tensor = torch.exp(-(B * frequency_grid_px**2) / divisor)
    return b_tensor


def Cs_envelope(
    spherical_aberration: float,  # in mm
    defocus: float,  # units in microns, positive for underfocus
    image_shape: tuple[int, int] | tuple[int, int, int],
    pixel_size: float,  # in angstroms
    rfft: bool = True,
    fftshift: bool = False,
    device: torch.device = None,
    voltage: float = 300,  # in kV
    alpha: float = 0.005,  # semiangle in mrad
) -> torch.Tensor:
    """
    Create a Cs envelope for a Fourier transform.

    Parameters
    ----------
    spherical_aberration: float
        The Cs value in mm
    defocus: float
        The defocus value in microns. Positive for underfocus.
    image_shape: tuple[int, ...]
        Shape of the real space the dft is from input image.
    pixel_size: float
        The pixel size of the image.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False).
    fftshift: bool
        Whether the input is fftshifted.
    device: torch.device
        Device to place tensors on.
    voltage: float
        The voltage of the microscope in kV.
    alpha: float
        The semiangle in mrad.

    Returns
    -------
    torch.Tensor
        Cs envelope
    """
    voltage *= 1e3  # kV -> V
    _lambda = (
        calculate_relativistic_electron_wavelength(voltage) * 1e10
    )  # wavelength meters -> angstroms
    Cs = spherical_aberration * 1e7  # mm -> angstroms
    defocus *= 1e4  # microns -> angstroms

    frequency_grid_px = (
        fftfreq_grid(
            image_shape=image_shape,
            rfft=rfft,
            fftshift=fftshift,
            norm=True,
            device=device,
        )
        / pixel_size
    )

    Cs_env = torch.exp(
        -(((torch.pi * (alpha / 1000)) / _lambda) ** 2)
        * (
            Cs * _lambda**3 * frequency_grid_px**3
            + _lambda * (defocus) * frequency_grid_px
        )
        ** 2
    )

    return Cs_env


def Cc_envelope(
    chromatic_aberration: float,  # in mm
    image_shape: tuple[int, int] | tuple[int, int, int],
    pixel_size: float,  # in angstroms
    rfft: bool = True,
    fftshift: bool = False,
    device: torch.device = None,
    voltage: float = 300,  # in kV
    energy_spread: float = 0.7,  # in eV
    deltaV_V: float = 0.06e-6,  #
    deltaI_I: float = 0.01e-6,  #
) -> torch.Tensor:
    """
    Create a Cc envelope for a Fourier transform.

    Parameters
    ----------
    chromatic_aberration: float
        The Cc value in mm
    image_shape: tuple[int, ...]
        Shape of the real space the dft is from input image.
    pixel_size: float
        The pixel size of the image.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False).
    fftshift: bool
        Whether the input is fftshifted.
    device: torch.device
        Device to place tensors on.
    voltage: float
        The voltage of the microscope in kV.
    energy_spread: float
        The FWHM of the energy spread in eV.
    deltaV_V: float
        The relative voltage spread.
    deltaI_I: float
        The relative current spread

    Returns
    -------
    torch.Tensor
        Cc envelope
    """
    voltage *= 1e3  # kV -> V
    _lambda = (
        calculate_relativistic_electron_wavelength(voltage) * 1e10
    )  # wavelength meters -> angstroms
    Cc = chromatic_aberration * 1e7  # mm -> angstroms

    frequency_grid_px = (
        fftfreq_grid(
            image_shape=image_shape,
            rfft=rfft,
            fftshift=fftshift,
            norm=True,
            device=device,
        )
        / pixel_size
    )

    focus_spread = Cc * (
        ((energy_spread / voltage) ** 2 + deltaV_V**2 + (2 * deltaI_I) ** 2) ** 0.5
    )
    Cc_env = torch.exp(
        -0.5 * ((torch.pi * _lambda * focus_spread * (frequency_grid_px) ** 2) ** 2)
    )

    return Cc_env



"""Bandpass filters using Fourier transform."""

import einops
import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def bandpass_filter(
    low: float,
    high: float,
    falloff: float,
    image_shape: tuple[int, int] | tuple[int, int, int],
    rfft: bool,
    fftshift: bool,
    device: torch.device = None,
) -> torch.Tensor:
    """
    Create a bandpass filter for the Fourier transform.

    Note that low and high are in frequency range 0-0.5 at Nyquist.

    Parameters
    ----------
    low: float
        Lower cutoff frequency.
    high: float
        Higher cutoff frequency.
    falloff: float
        Falloff of the filter.
    image_shape: tuple
        Shape of the image.
    rfft: bool
        If the FFT is a real FFT.
    fftshift: bool
        If the FFT is shifted.
    device: torch.device
        Device to use.

    Returns
    -------
    band_float: torch.Tensor
        Bandpass filter.
    """
    # get the frequency grid
    frequency_grid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=True,
        device=device,
    )
    # get the frequency band
    band = torch.logical_and(frequency_grid > low, frequency_grid <= high)
    band_float = band.float()

    # add cosine falloff
    band_with_falloff = torch.logical_and(
        frequency_grid > low - falloff, frequency_grid <= high + falloff
    )
    falloff_mask = torch.logical_and(band_with_falloff, ~band)
    cutoffs = torch.tensor([low, high], dtype=torch.float32, device=device)
    cutoffs = einops.rearrange(cutoffs, pattern="cutoffs -> cutoffs 1")
    distance = torch.abs(frequency_grid[falloff_mask] - cutoffs)
    distance = einops.reduce(distance, "cutoffs b -> b", reduction="min")
    softened_values = torch.cos((distance / falloff) * (torch.pi / 2))
    band_float[falloff_mask] = softened_values
    return band_float


def low_pass_filter(
    cutoff: float,
    falloff: float,
    image_shape: tuple[int, int] | tuple[int, int, int],
    rfft: bool,
    fftshift: bool,
    device: torch.device = None,
) -> torch.Tensor:
    """
    Create a low pass filter for the Fourier transform.

    Parameters
    ----------
    cutoff: float
        Cutoff frequency.
    falloff: float
        Falloff of the filter.
    image_shape: tuple
        Shape of the image.
    rfft: bool
        If the FFT is a real FFT.
    fftshift: bool
        If the FFT is shifted.
    device: torch.device
        Device to use.

    Returns
    -------
    filter: torch.Tensor
        Low pass filter.
    """
    filter_lp = bandpass_filter(
        low=0,
        high=cutoff,
        falloff=falloff,
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        device=device,
    )
    return filter_lp


def high_pass_filter(
    cutoff: float,
    falloff: float,
    image_shape: tuple[int, int] | tuple[int, int, int],
    rfft: bool,
    fftshift: bool,
    device: torch.device = None,
) -> torch.Tensor:
    """
    Create a high pass filter for the Fourier transform.

    Parameters
    ----------
    cutoff: float
        Cutoff frequency.
    falloff: float
        Falloff of the filter.
    image_shape: tuple
        Shape of the image.
    rfft: bool
        If the FFT is a real FFT.
    fftshift: bool
        If the FFT is shifted.
    device: torch.device
        Device to use.

    Returns
    -------
    filter: torch.Tensor
        High pass filter.
    """
    filter_hp = bandpass_filter(
        low=cutoff,
        high=1,
        falloff=falloff,
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        device=device,
    )
    return filter_hp



import torch

from torch_fourier_filter.envelopes import Cc_envelope, Cs_envelope, b_envelope


def test_bfactor_2d():
    # Generate an image
    image = torch.zeros((4, 4))
    image[1:, :] = 1
    # apply bfactor
    dft = torch.fft.rfftn(image, dim=(-2, -1))
    b_env = b_envelope(
        B=10,
        image_shape=image.shape,
        pixel_size=1,
        rfft=True,
        fftshift=False,
        device=dft.device,
    )
    bfactored_dft = dft * b_env
    result = torch.real(torch.fft.irfftn(bfactored_dft, dim=(-2, -1)))

    expected = torch.tensor(
        [
            [
                0.18851198,
                0.18851198,
                0.18851198,
                0.18851198,
            ],
            [0.88381535, 0.88381535, 0.88381535, 0.88381535],
            [1.0438573, 1.0438573, 1.0438573, 1.0438573],
            [0.88381535, 0.88381535, 0.88381535, 0.88381535],
        ]
    )
    assert torch.allclose(result, expected, atol=1e-3)


def test_Cc_envelope():
    chromatic_aberration = 2.0  # in mm
    image_shape = (4, 4)
    pixel_size = 1.0  # in angstroms
    rfft = True
    fftshift = False
    device = torch.device("cpu")
    voltage = 300  # in kV
    energy_spread = 0.7  # in eV
    deltaV_V = 0.06e-6
    deltaI_I = 0.01e-6

    Cc_env = Cc_envelope(
        chromatic_aberration,
        image_shape,
        pixel_size,
        rfft,
        fftshift,
        device,
        voltage,
        energy_spread,
        deltaV_V,
        deltaI_I,
    )

    assert isinstance(Cc_env, torch.Tensor), "Output is not a torch.Tensor"
    assert (
        Cc_env.min() >= 0.0 and Cc_env.max() <= 1.0
    ), "Values are out of expected range"

    # Test with real values
    image = torch.zeros((4, 4))
    image[1:, :] = 1
    dft = torch.fft.rfftn(image, dim=(-2, -1))

    Cc_dft = dft * Cc_env
    result = torch.real(torch.fft.irfftn(Cc_dft, dim=(-2, -1)))

    expected = torch.tensor(
        [
            [0.0654, 0.0654, 0.0654, 0.0654],
            [0.9427, 0.9427, 0.9427, 0.9427],
            [1.0493, 1.0493, 1.0493, 1.0493],
            [0.9427, 0.9427, 0.9427, 0.9427],
        ]
    )

    assert torch.allclose(result, expected, atol=1e-3)


def test_Cs_envelope():
    spherical_aberration = 2.0  # in mm
    defocus = 1.0  # in microns,
    image_shape = (4, 4)
    pixel_size = 1.0  # in angstroms
    rfft = True
    fftshift = False
    device = torch.device("cpu")
    voltage = 300  # in kV
    alpha = 0.005  # in mrad

    Cs_env = Cs_envelope(
        spherical_aberration,
        defocus,
        image_shape,
        pixel_size,
        rfft,
        fftshift,
        device,
        voltage,
        alpha,
    )

    assert isinstance(Cs_env, torch.Tensor), "Output is not a torch.Tensor"
    assert (
        Cs_env.min() >= 0.0 and Cs_env.max() <= 1.0
    ), "Values are out of expected range"

    # Test with real values
    image = torch.zeros((4, 4))
    image[1:, :] = 1
    dft = torch.fft.rfftn(image, dim=(-2, -1))

    Cs_dft = dft * Cs_env
    result = torch.real(torch.fft.irfftn(Cs_dft, dim=(-2, -1)))

    expected = torch.tensor(
        [
            [0.0030, 0.0030, 0.0030, 0.0030],
            [0.9978, 0.9978, 0.9978, 0.9978],
            [1.0013, 1.0013, 1.0013, 1.0013],
            [0.9978, 0.9978, 0.9978, 0.9978],
        ]
    )

    assert torch.allclose(result, expected, atol=1e-3)



import torch

from torch_fourier_filter.dose_weight import (
    critical_exposure,
    critical_exposure_bfactor,
    cumulative_dose_filter_3d,
)


def test_critical_exposure():
    fft_freq = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5])
    expected_output = torch.tensor([14.1383, 6.3823, 4.6287, 3.9365, 3.5869])
    output = critical_exposure(fft_freq)
    assert torch.allclose(
        output, expected_output, atol=1e-3
    ), "critical_exposure output mismatch"


def test_critical_exposure_bfactor():
    fft_freq = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5])
    bfac = 1.0
    expected_output = torch.tensor([200.0, 50.0, 22.222, 12.5, 8.0])
    output = critical_exposure_bfactor(fft_freq, bfac)
    assert torch.allclose(
        output, expected_output, atol=1e-3
    ), "critical_exposure_bfactor output mismatch"


def test_cumulative_dose_filter_3d():
    # Test parameters
    volume_shape = (32, 32, 32)
    start_exposure = 0.0
    end_exposure = 10.0
    pixel_size = 1.0
    crit_exposure_bfactor = -1
    rfft = True
    fftshift = False

    # Call the function
    dose_filter = cumulative_dose_filter_3d(
        volume_shape=volume_shape,
        pixel_size=pixel_size,
        start_exposure=start_exposure,
        end_exposure=end_exposure,
        crit_exposure_bfactor=crit_exposure_bfactor,
        rfft=rfft,
        fftshift=fftshift,
    )

    # Check if the values are within a reasonable range
    # TODO: Test these against known, static values rather than just range
    assert torch.all(dose_filter >= 0) and torch.all(
        dose_filter <= 1
    ), "Dose filter values out of range"

    # Test with different bfac values
    # TODO: Test these against known, static values rather than just range
    crit_bfac_values = [0.5, 1.0, 2.0]
    for bfac in crit_bfac_values:
        dose_filter = cumulative_dose_filter_3d(
            volume_shape=volume_shape,
            pixel_size=pixel_size,
            start_exposure=start_exposure,
            end_exposure=end_exposure,
            crit_exposure_bfactor=bfac,
            rfft=rfft,
            fftshift=fftshift,
        )
        assert torch.all(dose_filter >= 0) and torch.all(
            dose_filter <= 1
        ), f"Dose filter values out of range for bfac={bfac}"



"""CTF module for Fourier filtering."""

import einops
import torch
from scipy import constants as C
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def calculate_relativistic_electron_wavelength(energy: float) -> float:
    """Calculate the relativistic electron wavelength in SI units.

    For derivation see:
    1.  Kirkland, E. J. Advanced Computing in Electron Microscopy.
        (Springer International Publishing, 2020). doi:10.1007/978-3-030-33260-0.

    2.  https://en.wikipedia.org/wiki/Electron_diffraction#Relativistic_theory

    Parameters
    ----------
    energy: float
        acceleration potential in volts.

    Returns
    -------
    wavelength: float
        relativistic wavelength of the electron in meters.
    """
    h = C.Planck
    c = C.speed_of_light
    m0 = C.electron_mass
    e = C.elementary_charge
    V = energy
    eV = e * V

    numerator = h * c
    denominator = (eV * (2 * m0 * c**2 + eV)) ** 0.5
    return float(numerator / denominator)


def calculate_ctf_2d(
    defocus: float | torch.Tensor,
    astigmatism: float | torch.Tensor,
    astigmatism_angle: float | torch.Tensor,
    voltage: float,
    spherical_aberration: float | torch.Tensor,
    amplitude_contrast: float,
    b_factor: float | torch.Tensor,
    phase_shift: float | torch.Tensor,
    pixel_size: float,
    image_shape: tuple[int, int],
    rfft: bool,
    fftshift: bool,
) -> torch.Tensor:
    """Calculate the Contrast Transfer Function (CTF) for a 2D image.

    NOTE: The device of the input tensors is inferred from the `defocus` tensor.

    Parameters
    ----------
    defocus: float
        Defocus in micrometers, positive is underfocused.
        `(defocus_u + defocus_v) / 2`
    astigmatism: float
        Amount of astigmatism in micrometers.
        `(defocus_u - defocus_v) / 2`
    astigmatism_angle: float
        Angle of astigmatism in degrees. 0 places `defocus_u` along the y-axis.
    pixel_size: float
        Pixel size in Angströms per pixel (Å px⁻¹).
    voltage: float
        Acceleration voltage in kilovolts (kV).
    spherical_aberration: float
        Spherical aberration in millimeters (mm).
    amplitude_contrast: float
        Fraction of amplitude contrast (value in range [0, 1]).
    b_factor: float
        B-factor in square angstroms. Should be positive
    phase_shift: float
        Angle of phase shift applied to CTF in degrees.
    image_shape: Tuple[int, int]
        Shape of 2D images onto which CTF will be applied.
    rfft: bool
        Generate the CTF containing only the non-redundant half transform from a rfft.
        Only one of `rfft` and `fftshift` may be `True`.
    fftshift: bool
        Whether to apply fftshift on the resulting CTF images.

    Returns
    -------
    ctf: torch.Tensor
        The Contrast Transfer Function for the given parameters.
    """
    if isinstance(defocus, torch.Tensor):
        device = defocus.device
    else:
        device = torch.device("cpu")

    # to torch.Tensor
    cs_tensor = False
    defocus = torch.atleast_1d(
        torch.as_tensor(defocus, dtype=torch.float, device=device)
    )
    astigmatism = torch.atleast_1d(
        torch.as_tensor(astigmatism, dtype=torch.float, device=device)
    )
    astigmatism_angle = torch.atleast_1d(
        torch.as_tensor(astigmatism_angle, dtype=torch.float, device=device)
    )
    pixel_size = torch.atleast_1d(torch.as_tensor(pixel_size, device=device))
    voltage = torch.atleast_1d(
        torch.as_tensor(voltage, dtype=torch.float, device=device)
    )
    spherical_aberration = torch.atleast_1d(
        torch.as_tensor(spherical_aberration, dtype=torch.float, device=device)
    )
    if spherical_aberration.shape[0] > 1:
        cs_tensor = True
    image_shape = torch.as_tensor(image_shape, device=device)

    # Unit conversions
    defocus *= 1e4  # micrometers -> angstroms
    astigmatism *= 1e4  # micrometers -> angstroms
    astigmatism_angle *= C.pi / 180  # degrees -> radians
    voltage *= 1e3  # kV -> V
    spherical_aberration *= 1e7  # mm -> angstroms

    # derived quantities used in CTF calculation
    defocus_u = defocus + astigmatism
    defocus_v = defocus - astigmatism
    _lambda = (
        calculate_relativistic_electron_wavelength(voltage) * 1e10
    )  # meters -> angstroms
    k1 = -C.pi * _lambda
    k2 = C.pi / 2 * spherical_aberration * _lambda**3
    k3 = torch.deg2rad(torch.as_tensor(phase_shift, dtype=torch.float))
    k4 = -b_factor / 4
    amplitude_contrast = torch.as_tensor(amplitude_contrast, dtype=torch.float)
    k5 = torch.arctan(amplitude_contrast / torch.sqrt(1 - amplitude_contrast**2))

    # construct 2D frequency grids and rescale cycles / px -> cycles / Å
    fft_freq_grid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=False,
        device=device,
    )

    fft_freq_grid = fft_freq_grid / einops.rearrange(pixel_size, "b -> b 1 1 1")
    fftfreq_grid_squared = fft_freq_grid**2

    # Astigmatism
    #         Q = [[ sin, cos]
    #              [-sin, cos]]
    #         D = [[   u,   0]
    #              [   0,   v]]
    #         A = Q^T.D.Q = [[ Axx, Axy]
    #                        [ Ayx, Ayy]]
    #         Axx = cos^2 * u + sin^2 * v
    #         Ayy = sin^2 * u + cos^2 * v
    #         Axy = Ayx = cos * sin * (u - v)
    #         defocus = A.k.k^2 = Axx*x^2 + 2*Axy*x*y + Ayy*y^2

    c = torch.cos(astigmatism_angle)
    c2 = c**2
    s = torch.sin(astigmatism_angle)
    s2 = s**2

    yy2, xx2 = einops.rearrange(fftfreq_grid_squared, "b h w freq -> freq b h w")
    xy = einops.reduce(fft_freq_grid, "b h w freq -> b h w", reduction="prod")
    n4 = (
        einops.reduce(fftfreq_grid_squared, "b h w freq -> b h w", reduction="sum") ** 2
    )

    Axx = c2 * defocus_u + s2 * defocus_v
    Axy = c * s * (defocus_u - defocus_v)
    Ayy = s2 * defocus_u + c2 * defocus_v
    if cs_tensor:
        Axx_x2 = einops.rearrange(Axx, "... -> ... 1 1 1") * xx2
        Axy_xy = einops.rearrange(Axy, "... -> ... 1 1 1") * xy
        Ayy_y2 = einops.rearrange(Ayy, "... -> ... 1 1 1") * yy2
        k2 = einops.rearrange(k2, "... -> 1 1 ... 1 1")
    else:
        Ayy_y2 = einops.rearrange(Ayy, "... -> ... 1 1") * yy2
        Axx_x2 = einops.rearrange(Axx, "... -> ... 1 1") * xx2
        Axy_xy = einops.rearrange(Axy, "... -> ... 1 1") * xy

    # calculate ctf
    ctf = -torch.sin(k1 * (Axx_x2 + (2 * Axy_xy) + Ayy_y2) + k2 * n4 - k3 - k5)
    if k4 < 0:
        ctf *= torch.exp(k4 * n4)

    return ctf


def calculate_ctf_1d(
    defocus: float,
    voltage: float,
    spherical_aberration: float,
    amplitude_contrast: float,
    b_factor: float,
    phase_shift: float,
    pixel_size: float,
    n_samples: int,
    oversampling_factor: int,
) -> torch.Tensor:
    """
    Calculate the Contrast Transfer Function (CTF) for a 1D signal.

    Parameters
    ----------
    defocus: float
        Defocus in micrometers, positive is underfocused.
    pixel_size: float
        Pixel size in Angströms per pixel (Å px⁻¹).
    voltage: float
        Acceleration voltage in kilovolts (kV).
    spherical_aberration: float
        Spherical aberration in millimeters (mm).
    amplitude_contrast: float
        Fraction of amplitude contrast (value in range [0, 1]).
    b_factor: float
        B-factor in square angstroms. Should be positive
    phase_shift: float
        Angle of phase shift applied to CTF in degrees.
    n_samples: int
        Number of samples in CTF.
    oversampling_factor: int
        Factor by which to oversample the CTF.

    Returns
    -------
    ctf: torch.Tensor
        The Contrast Transfer Function for the given parameters
    """
    # to torch.Tensor and unit conversions
    defocus = torch.atleast_1d(torch.as_tensor(defocus, dtype=torch.float))
    defocus = defocus * 1e4  # micrometers -> angstroms
    defocus = einops.rearrange(defocus, "... -> ... 1")
    pixel_size = torch.atleast_1d(torch.as_tensor(pixel_size))
    voltage = torch.atleast_1d(torch.as_tensor(voltage, dtype=torch.float))
    voltage = voltage * 1e3  # kV -> V
    spherical_aberration = torch.atleast_1d(
        torch.as_tensor(spherical_aberration, dtype=torch.float)
    )
    spherical_aberration *= 1e7  # mm -> angstroms

    # derived quantities used in CTF calculation
    _lambda = (
        calculate_relativistic_electron_wavelength(voltage) * 1e10
    )  # meters -> angstroms
    k1 = -C.pi * _lambda
    k2 = C.pi / 2 * spherical_aberration * _lambda**3
    k3 = torch.deg2rad(torch.as_tensor(phase_shift, dtype=torch.float))
    k4 = -b_factor / 4
    amplitude_contrast = torch.as_tensor(amplitude_contrast, dtype=torch.float)
    k5 = torch.arctan(amplitude_contrast / torch.sqrt(1 - amplitude_contrast**2))

    # construct frequency vector and rescale cycles / px -> cycles / Å
    fftfreq_grid = torch.linspace(0, 0.5, steps=n_samples)  # (n_samples, )

    # oversampling...
    if oversampling_factor > 1:
        frequency_delta = 0.5 / (n_samples - 1)
        oversampled_frequency_delta = frequency_delta / oversampling_factor
        oversampled_interval_length = oversampled_frequency_delta * (
            oversampling_factor - 1
        )
        per_frequency_deltas = torch.linspace(
            0, oversampled_interval_length, steps=oversampling_factor
        )
        per_frequency_deltas -= oversampled_interval_length / 2
        per_frequency_deltas = einops.rearrange(per_frequency_deltas, "os -> os 1")
        fftfreq_grid = fftfreq_grid + per_frequency_deltas
        defocus = einops.rearrange(defocus, "... -> ... 1")  # oversampling dim

    fftfreq_grid = fftfreq_grid / pixel_size
    fftfreq_grid_squared = fftfreq_grid**2
    n4 = fftfreq_grid_squared**2

    # calculate ctf
    ctf = -torch.sin(k1 * fftfreq_grid_squared * defocus + k2 * n4 - k3 - k5)
    if k4 < 0:
        ctf *= torch.exp(k4 * n4)

    if oversampling_factor > 1:
        ctf = einops.reduce(ctf, "... os k -> ... k", reduction="mean")
    return ctf



"""Filters for image and volumes in pyTorch."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("torch-fourier-filter")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Josh Dickerson"
__email__ = "jdickerson@berkeley.edu"



"""Dose weighting functions for Fourier filtering."""

import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def critical_exposure(fft_freq: torch.Tensor) -> torch.Tensor:
    """
    Calculate the critical exposure using the Grant and Grigorieff 2015 formula.

    Args:
        fft_freq: The frequency grid of the Fourier transform.

    Returns
    -------
        The critical exposure for the given frequency grid
    """
    a = 0.245
    b = -1.665
    c = 2.81

    eps = 1e-10
    Ne = a * torch.pow(fft_freq.clamp(min=eps), b) + c
    return Ne


def critical_exposure_bfactor(fft_freq: torch.Tensor, bfactor: float) -> torch.Tensor:
    """
    Calculate the critical exposure using a user defined B-factor.

    Args:
        fft_freq: The frequency grid of the Fourier transform.
        bfactor: The B-factor to use.

    Returns
    -------
        The critical exposure for the given frequency grid
    """
    eps = 1e-10
    Ne = 2 / (bfactor * fft_freq.clamp(min=eps) ** 2)
    return Ne


def cumulative_dose_filter_3d(
    volume_shape: tuple[int, int, int] | tuple[int, int],
    pixel_size: float = 1,
    start_exposure: float = 0.0,
    end_exposure: float = 30.0,
    crit_exposure_bfactor: int | float = -1,
    rfft: bool = True,
    fftshift: bool = False,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Dose weight a 3D volume using Grant and Grigorieff 2015.

    Use integration to speed up.

    Parameters
    ----------
    volume_shape : tuple[int, int, int]
        The shape of the filter to calculate (real space). Rfft is
        automatically calculated from this.
    pixel_size : float
        The pixel size of the volume, in Angstrom.
    start_exposure : float
        The start exposure for dose weighting, in e-/A^2. Default is 0.0.
    end_exposure : float
        The end exposure for dose weighting, in e-/A^2. Default is 30.0.
    crit_exposure_bfactor : int | float
        The B factor for dose weighting based on critical exposure. If '-1',
        then use Grant and Grigorieff (2015) values.
    rfft : bool
        If the FFT is a real FFT.
    fftshift : bool
        If the FFT is shifted.
    device : torch.device
        The device to use for the calculation.

    Returns
    -------
    torch.Tensor
        The dose weighting filter.
    """
    # Get the frequency grid for 1 frame
    fft_freq_px = fftfreq_grid(
        image_shape=volume_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=True,
        device=device,
    )
    fft_freq_px /= pixel_size  # Convert to Angstrom^-1

    # Get the critical exposure for each frequency
    if crit_exposure_bfactor == -1:
        Ne = critical_exposure(fft_freq=fft_freq_px)
    elif crit_exposure_bfactor >= 0:
        Ne = critical_exposure_bfactor(
            fft_freq=fft_freq_px, bfactor=crit_exposure_bfactor
        )
    else:
        raise ValueError("B-factor must be positive or -1.")

    # Add small epsilon to prevent division by zero
    eps = 1e-10
    Ne = Ne.clamp(min=eps)

    return (
        2
        * Ne
        * (
            torch.exp((-0.5 * start_exposure) / Ne)
            - torch.exp((-0.5 * end_exposure) / Ne)
        )
        / end_exposure
    )



import torch

from torch_fourier_filter.whitening import (
    power_spectral_density,
    real_space_shape_from_dft_shape,
    whitening_filter,
)


def test_real_space_shape_from_dft_shape():
    """Basic tests for real_space_shape_from_dft_shape."""
    # Test 2D shape
    dft_shape = (64, 64)
    real_space_shape = real_space_shape_from_dft_shape(dft_shape, rfft=False)
    assert real_space_shape == (64, 64), "Real space shape incorrect for 2D input"

    # Test 2D shape with rfft
    dft_shape = (64, 33)
    real_space_shape = real_space_shape_from_dft_shape(dft_shape, rfft=True)
    assert real_space_shape == (
        64,
        64,
    ), "Real space shape incorrect for 2D input with rfft"

    # Test 3D shape
    dft_shape = (32, 64, 64)
    real_space_shape = real_space_shape_from_dft_shape(dft_shape, rfft=False)
    assert real_space_shape == (32, 64, 64), "Real space shape incorrect for 3D input"

    # Test 3D shape with rfft
    dft_shape = (32, 64, 33)
    real_space_shape = real_space_shape_from_dft_shape(dft_shape, rfft=True)
    assert real_space_shape == (
        32,
        64,
        64,
    ), "Real space shape incorrect for 3D input with rfft"


def test_power_spectral_density():
    """Basic tests for outputs of power_spectral_density."""
    img = torch.rand((64, 64))
    img_dft = torch.fft.rfftn(img)

    psd_1d = power_spectral_density(img_dft, rfft=True, fftshift=False)

    assert isinstance(psd_1d, torch.Tensor), "Output should be a tensor"
    assert psd_1d.ndim == 1, "Output should be 1D"


def test_whitening_filter():
    """Basic tests for outputs of whitening_filter."""
    img = torch.rand((64, 64))
    img_dft = torch.fft.rfftn(img)

    # Test for same output shape, etc.
    wf = whitening_filter(img_dft, rfft=False, fftshift=False)
    assert isinstance(wf, torch.Tensor), "Output should be a tensor"
    assert wf.shape == img_dft.shape, "Output shape should be the same as input"

    # Test for different output shape
    wf = whitening_filter(img_dft, rfft=False, fftshift=False, output_shape=(32, 32))
    assert isinstance(wf, torch.Tensor), "Output should be a tensor"
    assert wf.shape == (32, 32), "Output shape should be the same as input"

    # Test for different output shape
    wf = whitening_filter(img_dft, rfft=True, fftshift=False, output_shape=(32, 32))
    assert isinstance(wf, torch.Tensor), "Output should be a tensor"
    assert wf.shape == (32, 17), "Output shape should be the same as input"



"""Test functions in utils.py."""

import numpy as np
import pytest
import torch

from torch_fourier_filter.utils import (
    bin_1dim_with_lerp,
    curve_1dim_to_ndim,
    torch_interp,
)


def test_torch_interp():
    """Compare torch_interp to np.interp."""
    # Construct some random data
    x = torch.rand(1000) * 12 - 2
    xp = torch.linspace(0, 8, 200)
    fp = torch.sin(xp)

    # Test 1. Interpolation with other default values
    y_torch = torch_interp(x, xp, fp)
    y_numpy = np.interp(x.numpy(), xp.numpy(), fp.numpy())

    assert np.allclose(y_torch.numpy(), y_numpy, atol=1e-6)

    # Test 2. Interpolation with left and right values
    y_torch = torch_interp(x, xp, fp, left=10, right=10)
    y_numpy = np.interp(x.numpy(), xp.numpy(), fp.numpy(), left=10, right=10)

    assert np.allclose(y_torch.numpy(), y_numpy, atol=1e-6)

    # Test 3. NotImplementedError for period
    with pytest.raises(NotImplementedError):
        torch_interp(x, xp, fp, period=1.0)

    # Test 4. ValueError for different shapes
    with pytest.raises(ValueError):
        torch_interp(x, xp, fp[:-1])

    # Test 5. ValueError for multiple dimensions
    with pytest.raises(ValueError):
        torch_interp(x, xp.unsqueeze(0), fp)
    with pytest.raises(ValueError):
        torch_interp(x, xp, fp.unsqueeze(0))


def test_torch_interp_batched():
    """Basic test for interpolating across batched inputs."""
    # Construct some random data
    x = torch.rand(1000) * 12 - 2  # Grid is always constant
    xp0 = torch.linspace(0, 8, 200)
    xp1 = torch.linspace(0, 12, 200)
    xp2 = torch.linspace(4, 12, 200)
    fp0 = torch.sin(xp0)
    fp1 = torch.cos(xp1) + 1
    fp2 = torch.exp(xp2)

    # stack into batched inputs
    xp = torch.stack([xp0, xp1, xp2], dim=0)
    fp = torch.stack([fp0, fp1, fp2], dim=0)

    # Test 1. Interpolation with other default values
    y_torch = torch_interp(x, xp, fp)
    y0_numpy = np.interp(x.numpy(), xp0.numpy(), fp0.numpy())
    y1_numpy = np.interp(x.numpy(), xp1.numpy(), fp1.numpy())
    y2_numpy = np.interp(x.numpy(), xp2.numpy(), fp2.numpy())

    assert np.allclose(y_torch[0].numpy(), y0_numpy, atol=1e-6)
    assert np.allclose(y_torch[1].numpy(), y1_numpy, atol=1e-6)
    assert np.allclose(y_torch[2].numpy(), y2_numpy, atol=1e-6)


def test_curve_1dim_to_ndim():
    """Test converting a 1D curve at frequencies into a multi-dimensional grid."""

    def f(x):
        """Example function to construct data."""
        return torch.sin(2 * np.pi * x) + 2 * x

    def g(x):
        """Another example function to construct data."""
        return torch.exp(-(x**2)) * torch.cos(2 * np.pi * x)

    # Construct some example data
    frequencies = torch.linspace(0, 1, 1000)  # Needs to be highly sampled
    values = f(frequencies)

    # Construct a 2D grid
    x = torch.linspace(0, 0.5, 16)
    y = torch.linspace(0, 0.5, 16)
    xx, yy = torch.meshgrid(x, y, indexing="ij")
    frequency_grid = torch.sqrt(xx**2 + yy**2)

    # Test 1. Convert 1D curve to 2D grid
    filter_2d = curve_1dim_to_ndim(values, frequencies, frequency_grid)
    expected_2d = f(frequency_grid)

    assert torch.allclose(filter_2d, expected_2d, atol=1e-6)

    # Test 2. Batched inputs
    frequencies = torch.linspace(0, 1, 1000)  # Needs to be highly sampled
    values1 = f(frequencies)
    values2 = g(frequencies)
    values_batched = torch.stack([values1, values2], dim=0)

    filter_batched = curve_1dim_to_ndim(values_batched, frequencies, frequency_grid)
    expected1 = f(frequency_grid)
    expected2 = g(frequency_grid)
    expected_batched = torch.stack([expected1, expected2], dim=0)

    assert torch.allclose(filter_batched, expected_batched, atol=1e-6)

    # Test 3. ValueError for too many dimensions
    with pytest.raises(ValueError):
        curve_1dim_to_ndim(values_batched.unsqueeze(0), frequencies, frequency_grid)

    # Test 4. ValueError for different shapes
    with pytest.raises(ValueError):
        curve_1dim_to_ndim(values_batched, frequencies[:-1], frequency_grid)


def test_bin_1dim_with_lerp():
    # Test case 1: Basic functionality
    x = torch.tensor([0.5, 1.5, 2.5])
    y = torch.tensor([1.0, 2.0, 3.0])
    xp = torch.tensor([0.0, 1.0, 2.0, 3.0])
    expected_output = torch.tensor([0.5, 1.5, 2.5, 1.5])
    output = bin_1dim_with_lerp(x, y, xp, normalize_by_count=False)
    assert torch.allclose(
        output, expected_output
    ), f"Expected {expected_output}, but got {output}"

    # Test case 1b: Basic functionality with normalization
    x = torch.tensor([0.5, 1.5, 2.5])
    y = torch.tensor([1.0, 2.0, 3.0])
    xp = torch.tensor([0.0, 1.0, 2.0, 3.0])
    expected_output = torch.tensor([1.0, 1.5, 2.5, 3.0])
    output = bin_1dim_with_lerp(x, y, xp, normalize_by_count=True)
    assert torch.allclose(
        output, expected_output
    ), f"Expected {expected_output}, but got {output}"

    # Test case 2: x and y have different shapes
    x = torch.tensor([0.5, 1.5])
    y = torch.tensor([1.0, 2.0, 3.0])
    xp = torch.tensor([0.0, 1.0, 2.0, 3.0])
    with pytest.raises(ValueError, match="x and y must have the same shape."):
        bin_1dim_with_lerp(x, y, xp)

    # Test case 3: xp is not 1D
    x = torch.tensor([0.5, 1.5, 2.5])
    y = torch.tensor([1.0, 2.0, 3.0])
    xp = torch.tensor([[0.0, 1.0], [2.0, 3.0]])
    with pytest.raises(ValueError, match="xp must be 1D."):
        bin_1dim_with_lerp(x, y, xp)

    # Test case 4: Out of bounds values
    x = torch.tensor([-10, 0.0, 1.0, 10])
    y = torch.tensor([-1.0, 3.0, 9.0, -1.0])
    xp = torch.tensor([0.0, 1.0, 2.0, 3.0])
    expected_output = torch.tensor([3.0, 9.0, 0.0, 0.0])
    output = bin_1dim_with_lerp(x, y, xp)
    assert torch.allclose(
        output, expected_output
    ), f"Expected {expected_output}, but got {output}"



import torch

from torch_fourier_filter.mtf import (
    make_mtf_grid,
)


def test_make_mtf_grid():
    # Test parameters
    image_shape_2d = (64, 64)
    image_shape_3d = (32, 64, 64)
    mtf_frequencies = torch.linspace(0, 0.5, 10)
    mtf_amplitudes = torch.linspace(1, 0, 10)

    mtf_grid_2d = make_mtf_grid(
        image_shape_2d, mtf_frequencies, mtf_amplitudes, rfft=True, fftshift=False
    )
    mtf_grid_3d = make_mtf_grid(
        image_shape_3d, mtf_frequencies, mtf_amplitudes, rfft=True, fftshift=False
    )

    # Test if the MTF grid values are within the expected range
    assert torch.all(mtf_grid_2d >= 0) and torch.all(
        mtf_grid_2d <= 1
    ), "MTF grid values out of range for 2D input"
    assert torch.all(mtf_grid_3d >= 0) and torch.all(
        mtf_grid_3d <= 1
    ), "MTF grid values out of range for 3D input"



import torch

from torch_fourier_filter.bandpass import (
    bandpass_filter,
)


def test_bandpass_filter():
    freqs = torch.fft.fftfreq(20)
    in_band_idx = torch.logical_and(freqs >= 0.2, freqs <= 0.4)
    lower_idx = torch.logical_and(freqs >= 0.1, freqs <= 0.2)
    upper_idx = freqs > 0.4
    lower_falloff = torch.cos((torch.pi / 2) * ((freqs[lower_idx] - 0.2) / 0.1))
    upper_falloff = torch.cos((torch.pi / 2) * ((freqs[upper_idx] - 0.4) / 0.1))

    bp_filter = bandpass_filter(
        low=0.2,
        high=0.4,
        falloff=0.1,
        image_shape=(20, 1),
        rfft=False,
        fftshift=False,
        device=freqs.device,
    )

    assert torch.all(bp_filter[in_band_idx] == 1)
    assert torch.allclose(bp_filter[lower_idx], lower_falloff.view((-1, 1)), atol=1e-6)
    assert torch.allclose(bp_filter[upper_idx], upper_falloff.view((-1, 1)), atol=1e-6)



import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid

from torch_fourier_filter.phase_randomize import phase_randomize


def test_phase_randomize():
    # Test parameters
    image_shape_2d = (64, 64)
    image_shape_3d = (32, 64, 64)
    cuton = 0.1
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Generate random 2D and 3D DFTs
    dft_2d = torch.fft.fft2(torch.rand(image_shape_2d, device=device))
    dft_3d = torch.fft.fftn(torch.rand(image_shape_3d, device=device))

    # Test 2D phase randomization preserves magnitudes
    randomized_2d = phase_randomize(dft_2d, image_shape_2d, cuton=cuton, device=device)
    assert torch.allclose(
        torch.abs(dft_2d), torch.abs(randomized_2d)
    ), "Magnitude spectrum not preserved for 2D input"

    # Test 3D phase randomization preserves magnitudes
    randomized_3d = phase_randomize(dft_3d, image_shape_3d, cuton=cuton, device=device)
    assert torch.allclose(
        torch.abs(dft_3d), torch.abs(randomized_3d)
    ), "Magnitude spectrum not preserved for 3D input"

    # Test rfft preserves magnitudes
    dft_2d_rfft = torch.fft.rfft2(torch.rand(image_shape_2d, device=device))
    randomized_2d_rfft = phase_randomize(
        dft_2d_rfft, image_shape_2d, rfft=True, cuton=cuton, device=device
    )
    assert torch.allclose(
        torch.abs(dft_2d_rfft), torch.abs(randomized_2d_rfft)
    ), "Magnitude spectrum not preserved for rfft input"

    # Ensure phases are randomized only above cuton
    freq_grid_2d = fftfreq_grid(
        image_shape_2d, rfft=False, fftshift=False, norm=True, device=device
    )
    freq_mask_2d = freq_grid_2d > cuton
    original_phases_2d = torch.angle(dft_2d)
    randomized_phases_2d = torch.angle(randomized_2d)
    assert torch.allclose(
        original_phases_2d[~freq_mask_2d], randomized_phases_2d[~freq_mask_2d]
    ), "Phases not preserved below cuton for 2D input"
    assert not torch.allclose(
        original_phases_2d[freq_mask_2d], randomized_phases_2d[freq_mask_2d]
    ), "Phases not randomized above cuton for 2D input"

    freq_grid_3d = fftfreq_grid(
        image_shape_3d, rfft=False, fftshift=False, norm=True, device=device
    )
    freq_mask_3d = freq_grid_3d > cuton
    original_phases_3d = torch.angle(dft_3d)
    randomized_phases_3d = torch.angle(randomized_3d)
    assert torch.allclose(
        original_phases_3d[~freq_mask_3d], randomized_phases_3d[~freq_mask_3d]
    ), "Phases not preserved below cuton for 3D input"
    assert not torch.allclose(
        original_phases_3d[freq_mask_3d], randomized_phases_3d[freq_mask_3d]
    ), "Phases not randomized above cuton for 3D input"



import torch

from torch_fourier_filter.ctf import (
    calculate_ctf_1d,
    calculate_relativistic_electron_wavelength,
)


def test_1d_ctf_single():
    result = calculate_ctf_1d(
        defocus=1.5,
        pixel_size=8,
        voltage=300,
        spherical_aberration=2.7,
        amplitude_contrast=0.1,
        b_factor=0,
        phase_shift=0,
        n_samples=10,
        oversampling_factor=3,
    )
    expected = torch.tensor(
        [
            0.1033,
            0.1476,
            0.2784,
            0.4835,
            0.7271,
            0.9327,
            0.9794,
            0.7389,
            0.1736,
            -0.5358,
        ]
    )
    assert torch.allclose(result[0], expected, atol=1e-4)


def test_1d_ctf_batch_defocus():
    result = calculate_ctf_1d(
        defocus=[1.5, 2.5],
        pixel_size=8,
        voltage=300,
        spherical_aberration=2.7,
        amplitude_contrast=0.1,
        b_factor=0,
        phase_shift=0,
        n_samples=10,
        oversampling_factor=1,
    )
    print(result)
    expected = torch.tensor(
        [
            [
                0.1000,
                0.1444,
                0.2755,
                0.4819,
                0.7283,
                0.9385,
                0.9903,
                0.7519,
                0.1801,
                -0.5461,
            ],
            [
                0.1000,
                0.1738,
                0.3880,
                0.6970,
                0.9617,
                0.9237,
                0.3503,
                -0.5734,
                -0.9877,
                -0.1474,
            ],
        ]
    )
    assert torch.allclose(result, expected, atol=1e-4)


def test_calculate_relativistic_electron_wavelength():
    """Check function matches expected value from literature.

    De Graef, Marc (2003-03-27).
    Introduction to Conventional Transmission Electron Microscopy.
    Cambridge University Press. doi:10.1017/cbo9780511615092
    """
    result = calculate_relativistic_electron_wavelength(300e3)
    expected = 1.969e-12
    assert abs(result - expected) < 1e-15



"""Noise whitening functions."""

from typing import Optional

import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid

from torch_fourier_filter.utils import (
    _handle_dim,
    bin_1dim_with_lerp,
    curve_1dim_to_ndim,
)


def calculate_num_freq_bins(image_shape: tuple[int, ...]) -> int:
    """Calculate the number of frequency bins for a power spectrum given image shape.

    Parameters
    ----------
    image_shape: tuple[int, ...]
        Shape of the image.

    Returns
    -------
    int
        Number of frequency bins.
    """
    largest_dim = max(image_shape)

    return int((largest_dim // 2 + 1) * len(image_shape) ** 0.5 + 1)


def real_space_shape_from_dft_shape(
    dft_shape: tuple[int, ...], rfft: bool
) -> tuple[int, ...]:
    """Calculate the real-space shape from a DFT shape and rfft argument.

    Parameters
    ----------
    dft_shape: tuple[int, ...]
        Shape of the DFT.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False). The default is
        True.

    Returns
    -------
    tuple[int, ...]
        Shape of the real-space image.
    """
    real_space_shape = list(dft_shape)
    if rfft:
        last_dim = 2 * (real_space_shape[-1] - 1)
        real_space_shape[-1] = last_dim

    return tuple(real_space_shape)


def power_spectral_density(
    image_dft: torch.Tensor,
    rfft: bool = True,
    fftshift: bool = False,
    dim: int | tuple[int, ...] = (-2, -1),
    num_freq_bins: Optional[int] = None,
    max_freq: Optional[float] = None,
    do_power_spectrum: Optional[bool] = True,
) -> torch.Tensor:
    """Calculates the power spectral density of an image in 1-dimension.

    Note that the power spectrum can either be calculated over the intensities or
    amplitudes of the Fourier components; the parameter `do_power_spectrum` controls
    this.

    Parameters
    ----------
    image_dft: torch.Tensor
        The DFT of the images/volumes.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False).
    fftshift: bool
        Whether the input is fftshifted.
    dim: int | tuple[int, ...], optional
        Which dimensions to calculate the power spectrum over. The default is (-2, -1)
        which corresponds to the spatial dimensions of an image.
    num_freq_bins: int, optional
        Number of frequency bins to calculate the power spectrum over. If None, the
        the number is inferred from the input shape automatically.
    max_freq: float, optional
        Optional maximum frequency to calculate the power spectrum to, in terms of the
        Nyquist frequency. The default is None and corresponds to going past the Nyquist
        frequency into the corner of Fourier space. For example, a 2D input will have
        frequencies up to sqrt(2)/2. Frequencies above this are set to zero.
    do_power_spectrum: bool, optional
        Weather to use the power (intensity) spectrum or the amplitude spectrum. If
        True, the power spectrum is calculated, otherwise the amplitude spectrum is
        calculated. The default is True.

    Returns
    -------
    torch.Tensor
        The power spectral density of the input image. Includes batch dimensions, if
        present.

    Raises
    ------
    ValueError
        Currently *does not* support batching. Raises error if the input image is larger
        than the number of dimensions.
    """
    dim = _handle_dim(dim, image_dft.ndim)
    shape_over_dim = tuple(image_dft.shape[d] for d in dim)

    real_space_shape = real_space_shape_from_dft_shape(
        dft_shape=shape_over_dim, rfft=rfft
    )

    # Construct 1-dimensional grid of frequencies for binning
    if num_freq_bins is None:
        num_freq_bins = calculate_num_freq_bins(shape_over_dim)
    binning_freqs = torch.linspace(
        start=0.0,
        end=(len(dim) ** 0.5) / 2,  # corner of Fourier space
        steps=num_freq_bins,
        device=image_dft.device,
    )

    # Calculate the amplitude (or intensity) of all Fourier components
    power_spectrum = torch.abs(image_dft)
    if do_power_spectrum:
        power_spectrum = power_spectrum**2

    # Construct FFT frequency grid
    freq_grid = fftfreq_grid(
        image_shape=real_space_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=True,
        device=image_dft.device,
    )

    # Bin the power spectrum into 1-dimension
    power_spectrum_1d = bin_1dim_with_lerp(
        x=freq_grid,
        y=power_spectrum,
        xp=binning_freqs,
        normalize_by_count=True,
    )

    if do_power_spectrum:
        power_spectrum_1d = torch.sqrt(power_spectrum_1d)

    # Set frequencies above max_freq to zero, if specified
    if max_freq is not None:
        power_spectrum_1d = torch.where(
            binning_freqs <= max_freq,
            power_spectrum_1d,
            torch.zeros_like(power_spectrum_1d),
        )

    return power_spectrum_1d


def whitening_filter(
    image_dft: torch.Tensor,
    rfft: bool = True,
    fftshift: bool = False,
    dim: int | tuple[int, ...] = (-2, -1),
    num_freq_bins: Optional[int] = None,
    max_freq: Optional[float] = None,
    do_power_spectrum: bool = True,
    output_shape: Optional[tuple[int, ...]] = None,
    output_rfft: Optional[bool] = None,
    output_fftshift: Optional[bool] = None,
) -> torch.Tensor:
    """Create a whitening filter the discrete Fourier transform of an input.

    Parameters
    ----------
    image_dft: torch.Tensor
        The DFT of the images/volumes.
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False).
    fftshift: bool
        Whether the input is fftshifted.
    dim: int | tuple[int, ...], optional
        Which dimensions to calculate the power spectrum over. The default is (-2, -1)
        which corresponds to the spatial dimensions of an image.
    num_freq_bins: int, optional
        Number of frequency bins to calculate the power spectrum over. If None, the
        the number is inferred from the input shape automatically.
    max_freq: float, optional
        Maximum frequency to calculate the whitening filter to, in terms of the
        Nyquist frequency. The default is None and corresponds to going past the Nyquist
        frequency into the corner of Fourier space. For example, a 2D input will have
        frequencies up to sqrt(2)/2. Frequencies above this are set to one in the
        returned whitening filter.
    do_power_spectrum: bool, optional
        Weather to use the power (intensity) spectrum or the amplitude spectrum. If
        True, the power spectrum is calculated, otherwise the amplitude spectrum is
        calculated. The default is True.
    output_shape: tuple[int, ...], optional
        The shape of the output filter. If None, then the shape is the same as
        `image_dft`. The default is None. NOTE: This shape is in real space.
    output_rfft: bool, optional
        Whether to return the filter in rfft form. If None, then the same as `rfft`.
        The default is None.
    output_fftshift: bool, optional
        Whether to return the filter fftshifted. If None, then the same as `fftshift`.
        The default is None.

    Returns
    -------
    torch.Tensor
        The whitening filter in Fourier space.
    """
    dim = _handle_dim(dim, image_dft.ndim)

    power_spec_1d = power_spectral_density(
        image_dft=image_dft,
        rfft=rfft,
        fftshift=fftshift,
        dim=dim,
        num_freq_bins=num_freq_bins,
        do_power_spectrum=do_power_spectrum,
    )

    # Create the filter from the power spectrum
    psd_zeros_mask = power_spec_1d == 0
    whitening_filter_1d = torch.where(
        psd_zeros_mask, torch.zeros_like(power_spec_1d), 1 / power_spec_1d
    )

    # Calculate the last frequency value
    last_freq = (len(dim) ** 0.5) / 2

    # Generate a 1D frequency grid
    frequency_1d = torch.linspace(0, last_freq, steps=len(whitening_filter_1d))

    # Create a mask for frequencies above the max frequency, if specified
    if max_freq is not None:
        above_freq_mask = frequency_1d > max_freq
        whitening_filter_1d[above_freq_mask] = 1.0
    else:
        above_freq_mask = torch.zeros_like(whitening_filter_1d, dtype=torch.bool)

    # Normalize the whitening filter by the maximum value of the valid frequencies
    max_valid_value = whitening_filter_1d[~above_freq_mask].max()
    whitening_filter_1d /= max_valid_value

    # Set the values above the max frequency to 1
    whitening_filter_1d = torch.where(
        above_freq_mask, torch.ones_like(whitening_filter_1d), whitening_filter_1d
    )

    # Construct FFT frequency grid
    if output_shape is None:
        output_shape = real_space_shape_from_dft_shape(image_dft.shape, rfft=rfft)

    if output_rfft is None:
        output_rfft = rfft

    if output_fftshift is None:
        output_fftshift = fftshift

    # real_space_shape = real_space_shape_from_dft_shape(
    #     dft_shape=output_shape, rfft=output_rfft
    # )

    freq_grid = fftfreq_grid(
        image_shape=output_shape,
        rfft=output_rfft,
        fftshift=output_fftshift,
        norm=True,
        device=image_dft.device,
    )

    # Interpolate the filter to the frequency grid
    whitening_filter_ndim = curve_1dim_to_ndim(
        frequency_1d=frequency_1d,
        values_1d=whitening_filter_1d,
        frequency_grid=freq_grid,
        fill_lower=1.0,  # Fill oob areas with ones (no scaling)
        fill_upper=1.0,
    )

    return whitening_filter_ndim


# def gaussian_smoothing(
#     tensor: torch.Tensor,
#     dim: int | tuple[int, ...] = -1,
#     kernel_size: int = 5,
#     sigma: float = 1.0,
# ) -> torch.Tensor:
#     """
#     Apply Gaussian smoothing over specified dimensions of a tensor.

#     Parameters
#     ----------
#     tensor: torch.Tensor
#         The input tensor to be smoothed.
#     dim: int | tuple[int, ...]
#         Dimensions over which to apply smoothing. Can be a single int or tuple of
#         ints. Negative dimensions are indexed from the end.
#     kernel_size: int
#         The size of the Gaussian kernel.
#     sigma: float
#         The standard deviation of the Gaussian kernel.

#     Returns
#     -------
#     torch.Tensor
#         The smoothed tensor with same shape as input.
#     """
#     # Convert single dim to tuple
#     if isinstance(dim, int):
#         dim = (dim,)

#     # Convert negative dims to positive
#     dim = tuple(d if d >= 0 else tensor.dim() + d for d in dim)

#     # Validate dimensions
#     if not all(0 <= d < tensor.dim() for d in dim):
#         raise ValueError(
#             f"Invalid dimensions {dim} for tensor of rank {tensor.dim()}"
#          )
#     if len(dim) > 2:
#         raise ValueError("Gaussian smoothing only supports 1D or 2D operations")

#     # Create coordinate grid for kernel
#     x = torch.arange(
#         -kernel_size // 2 + 1,
#         kernel_size // 2 + 1,
#         dtype=tensor.dtype,
#         device=tensor.device,
#     )

#     if len(dim) == 1:  # 1D smoothing
#         kernel = torch.exp(-0.5 * (x / sigma) ** 2)
#         kernel = kernel / kernel.sum()

#         # Create conv kernel with singleton dimensions
#         shape = [1] * (tensor.dim())
#         shape[dim[0]] = kernel_size
#         kernel = kernel.view(*shape)

#         # Apply 1D convolution
#         tensor = einops.rearrange(tensor, "... n -> ... 1 1 n")
#         kernel = einops.rearrange(kernel, "... n -> ... 1 1 n")
#         smoothed_tensor = F.conv1d(tensor, kernel, padding=kernel_size // 2)
#         return einops.rearrange(smoothed_tensor, "... 1 1 n -> ... n")

#     else:  # 2D smoothing
#         x, y = torch.meshgrid(x, x, indexing="ij")
#         kernel = torch.exp(-0.5 * ((x / sigma) ** 2 + (y / sigma) ** 2))
#         kernel = kernel / kernel.sum()

#         # Create conv kernel with singleton dimensions
#         shape = [1] * (tensor.dim())
#         shape[dim[0]] = kernel_size
#         shape[dim[1]] = kernel_size
#         kernel = kernel.view(*shape)
#         kernel = einops.rearrange(kernel, "... h w -> ... 1 1 h w")
#         # Apply 2D convolution
#         tensor = einops.rearrange(tensor, "... h w -> ... 1 1 h w")
#         smoothed_tensor = F.conv2d(
#             tensor, kernel, padding=kernel_size // 2, stride=(1, 1)
#         )
#         return einops.rearrange(smoothed_tensor, "... 1 1 h w -> ... h w")



"""Phase randomization function."""

import torch
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def phase_randomize(
    dft: torch.Tensor,
    image_shape: tuple[int, int] | tuple[int, int, int],
    rfft: bool = False,
    cuton: float = 0,
    fftshift: bool = False,
    device: torch.device = None,
) -> torch.Tensor:
    """Phase randomize a 2/3D Fourier transform while preserving magnitude spectrum.

    Parameters
    ----------
    dft: torch.Tensor
        Complex tensor containing 2D Fourier transform(s). Can be batched with shape
        (batch, h, w) or unbatched (h, w).
    image_shape: tuple[int, ...]
        Shape of the input image
    rfft: bool
        Whether the input is from an rfft (True) or full fft (False)
    cuton: float
        Fraction of Nyquist frequency to cut on
    fftshift: bool
        Whether the input is fftshifted
    device: torch.device
        Device to place tensors on

    Returns
    -------
    torch.Tensor
        Phase randomized version of input with same shape and dtype
    """
    # Get magnitude spectrum
    magnitudes = torch.abs(dft)

    # Create frequency grid
    # image_shape = dft.shape[-2:]
    freq_grid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshift,
        norm=True,
        device=device,
    )

    # Create mask for frequencies above cuton
    freq_mask = freq_grid > cuton

    # Generate random phases between -π and π only where freq > cuton
    random_phases = torch.zeros_like(dft, dtype=torch.float32)
    random_phases[..., freq_mask] = (
        torch.rand(freq_mask.sum(), device=dft.device) * (2 * torch.pi) - torch.pi
    )

    # Convert to complex numbers (e^(iθ))
    phase_factors = torch.complex(torch.cos(random_phases), torch.sin(random_phases))

    # Keep original phases where freq <= cuton
    original_phases = torch.angle(dft)
    original_phase_factors = torch.complex(
        torch.cos(original_phases), torch.sin(original_phases)
    )
    phase_factors[..., ~freq_mask] = original_phase_factors[..., ~freq_mask]

    # Combine with original magnitudes
    randomized = magnitudes * phase_factors

    return randomized



"""DFT utilities but currently only contains rotational averaging functions."""

from collections.abc import Sequence

import einops
import torch
import torch.nn.functional as F
from torch_grid_utils.coordinate_grid import coordinate_grid
from torch_grid_utils.fftfreq_grid import fftfreq_grid


def rotational_average_dft_2d(
    dft: torch.Tensor,
    image_shape: tuple[int, ...],
    rfft: bool = False,
    fftshifted: bool = False,
    return_1d_average: bool = True,
) -> tuple[torch.Tensor, torch.Tensor]:  # rotational_average, frequency_bins
    """
    Calculate the rotational average of a 2D DFT.

    Parameters
    ----------
    dft : torch.Tensor
        Complex tensor containing 2D Fourier transform(s). Can be batched with shape
        (batch, h, w) or unbatched (h, w).
    image_shape : tuple[int, ...]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted
    return_1d_average : bool
        If true, return a 1D rotational average and frequency bins, otherwise
        return a 2D average.

    Returns
    -------
    torch.Tensor
        Rotational average of the input DFT
    """
    # calculate the number of frequency bins
    h, w = image_shape[-2:]
    n_bins = min((d // 2) + 1 for d in (h, w))

    # split data into frequency bins
    frequency_bins = _frequency_bin_centers(n_bins, device=dft.device)
    shell_data = _split_into_frequency_bins_2d(
        dft, n_bins=n_bins, image_shape=(h, w), rfft=rfft, fftshifted=fftshifted
    )

    # calculate mean over each shell
    mean_per_shell = [
        einops.reduce(shell, "... shell -> ...", reduction="mean")
        for shell in shell_data
    ]
    rotational_average = einops.rearrange(mean_per_shell, "shells ... -> ... shells")
    if not return_1d_average:
        if len(dft.shape) > len(image_shape):
            image_shape = (*dft.shape[:-2], *image_shape[-2:])
        rotational_average = _1d_to_rotational_average_2d_dft(
            values=rotational_average,
            image_shape=image_shape,
            rfft=rfft,
            fftshifted=fftshifted,
        )
        frequency_bins = _1d_to_rotational_average_2d_dft(
            values=frequency_bins,
            image_shape=image_shape,
            rfft=rfft,
            fftshifted=fftshifted,
        )
    return rotational_average, frequency_bins


def rotational_average_dft_3d(
    dft: torch.Tensor,
    image_shape: tuple[int, ...],
    rfft: bool = False,
    fftshifted: bool = False,
    return_1d_average: bool = True,
) -> tuple[torch.Tensor, torch.Tensor]:  # rotational_average, frequency_bins
    """
    Calculate the rotational average of a 3D DFT.

    Parameters
    ----------
    dft : torch.Tensor
        Complex tensor containing 3D Fourier transform(s). Can be batched with shape
        (batch, d, h, w) or unbatched (d, h, w).
    image_shape : tuple[int, ...]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted
    return_1d_average : bool
        If true, return a 1D rotational average and frequency bins, otherwise
        return a 3D average.

    Returns
    -------
    torch.Tensor
        Rotational average of the input DFT
    """
    # calculate the number of frequency bins
    d, h, w = image_shape[-3:]
    n_bins = min((i // 2) + 1 for i in (h, w))

    # split data into frequency bins
    frequency_bins = _frequency_bin_centers(n_bins, device=dft.device)
    shell_data = _split_into_frequency_bins_3d(
        dft, n_bins=n_bins, image_shape=(d, h, w), rfft=rfft, fftshifted=fftshifted
    )

    # calculate mean over each shell
    mean_per_shell = [
        einops.reduce(shell, "... shell -> ...", reduction="mean")
        for shell in shell_data
    ]
    rotational_average = einops.rearrange(mean_per_shell, "shells ... -> ... shells")
    if not return_1d_average:
        if len(dft.shape) > len(image_shape):
            image_shape = (*dft.shape[:-3], *image_shape[-3:])
        rotational_average = _1d_to_rotational_average_3d_dft(
            values=rotational_average,
            image_shape=image_shape,
            rfft=rfft,
            fftshifted=fftshifted,
        )
        frequency_bins = _1d_to_rotational_average_3d_dft(
            values=frequency_bins,
            image_shape=image_shape,
            rfft=rfft,
            fftshifted=fftshifted,
        )
    return rotational_average, frequency_bins


def _find_shell_indices_1d(
    values: torch.Tensor, split_values: torch.Tensor
) -> list[torch.Tensor]:
    """
    Find indices which index to give values either side of split points.

    Parameters
    ----------
    values : torch.Tensor
        The values to split
    split_values : torch.Tensor
        The values to split at

    Returns
    -------
    list[torch.Tensor]
        List of tensors containing the indices of values in each shell
    """
    sorted_vals, sort_idx = torch.sort(values, descending=False)
    split_idx = torch.searchsorted(sorted_vals, split_values)
    # tensor_split requires the split_idx to live on cpu
    return list(torch.tensor_split(sort_idx, split_idx.to("cpu")))


def _find_shell_indices_2d(
    values: torch.Tensor, split_values: torch.Tensor
) -> list[torch.Tensor]:
    """
    Find 2D indices which index to give values either side of split values.

    Parameters
    ----------
    values : torch.Tensor
        The values to split
    split_values : torch.Tensor
        The values to split at

    Returns
    -------
    list[torch.Tensor]
        List of tensors containing the indices of values in each shell
    """
    idx_2d = coordinate_grid(values.shape[-2:]).long()
    values = einops.rearrange(values, "h w -> (h w)")
    idx_2d = einops.rearrange(idx_2d, "h w idx -> (h w) idx")
    sorted_vals, sort_idx = torch.sort(values, descending=False)
    split_idx = torch.searchsorted(sorted_vals, split_values)
    # tensor_split requires the split_idx to live on cpu
    return list(torch.tensor_split(idx_2d[sort_idx], split_idx.to("cpu")))


def _find_shell_indices_3d(
    values: torch.Tensor, split_values: torch.Tensor
) -> list[torch.Tensor]:
    """
    Find 3D indices which index to give values either side of split values.

    Parameters
    ----------
    values : torch.Tensor
        The values to split
    split_values : torch.Tensor
        The values to split at

    Returns
    -------
    list[torch.Tensor]
        List of tensors containing the indices of values in each shell
    """
    idx_3d = coordinate_grid(values.shape[-3:]).long()
    values = einops.rearrange(values, "d h w -> (d h w)")
    idx_3d = einops.rearrange(idx_3d, "d h w idx -> (d h w) idx")
    sorted_vals, sort_idx = torch.sort(values, descending=False)
    split_idx = torch.searchsorted(sorted_vals, split_values)
    # tensor_split requires the split_idx to live on cpu
    return list(torch.tensor_split(idx_3d[sort_idx], split_idx.to("cpu")))


def _split_into_frequency_bins_2d(
    dft: torch.Tensor,
    n_bins: int,
    image_shape: tuple[int, int],
    rfft: bool = False,
    fftshifted: bool = False,
) -> list[torch.Tensor]:
    """
    Split a 2D DFT into frequency bins.

    Parameters
    ----------
    dft : torch.Tensor
        Complex tensor containing 2D Fourier transform(s). Can be batched with shape
        (batch, h, w) or unbatched (h, w).
    n_bins : int
        Number of frequency bins to split the DFT into
    image_shape : tuple[int, int]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted

    Returns
    -------
    list[torch.Tensor]
        List of tensors containing the DFT values in each frequency bin
    """
    frequency_grid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshifted,
        norm=True,
        device=dft.device,
    )
    frequency_grid = einops.rearrange(frequency_grid, "h w -> (h w)")
    shell_borders = _frequency_bin_split_values(n_bins)
    shell_indices = _find_shell_indices_1d(frequency_grid, split_values=shell_borders)
    dft = einops.rearrange(dft, "... h w -> ... (h w)")
    shells = [dft[..., shell_idx] for shell_idx in shell_indices]
    return shells[:-1]


def _split_into_frequency_bins_3d(
    dft: torch.Tensor,
    n_bins: int,
    image_shape: tuple[int, int, int],
    rfft: bool = False,
    fftshifted: bool = False,
) -> list[torch.Tensor]:
    """
    Split a 3D DFT into frequency bins.

    Parameters
    ----------
    dft : torch.Tensor
        Complex tensor containing 3D Fourier transform(s). Can be batched with shape
        (batch, d, h, w) or unbatched (d, h, w).
    n_bins : int
        Number of frequency bins to split the DFT into
    image_shape : tuple[int, int, int]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted

    Returns
    -------
    list[torch.Tensor]
        List of tensors containing the DFT values in each frequency bin
    """
    frequency_grid = fftfreq_grid(
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshifted,
        norm=True,
        device=dft.device,
    )
    frequency_grid = einops.rearrange(frequency_grid, "d h w -> (d h w)")
    shell_borders = _frequency_bin_split_values(n_bins)
    shell_indices = _find_shell_indices_1d(frequency_grid, split_values=shell_borders)
    dft = einops.rearrange(dft, "... d h w -> ... (d h w)")
    shells = [dft[..., shell_idx] for shell_idx in shell_indices]
    return shells[:-1]


def _1d_to_rotational_average_2d_dft(
    values: torch.Tensor,
    image_shape: tuple[int, ...],
    rfft: bool = False,
    fftshifted: bool = True,
) -> torch.Tensor:
    """
    Convert a 1D array of values to a 2D array of values in frequency bins.

    Parameters
    ----------
    values : torch.Tensor
        1D tensor of values to convert
    image_shape : tuple[int, ...]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted

    Returns
    -------
    torch.Tensor
        2D tensor of values in frequency bins
    """
    # construct output tensor
    h, w = image_shape[-2:]
    h, w = rfft_shape((h, w)) if rfft is True else (h, w)
    result_shape = (*image_shape[:-2], h, w)
    average_2d = torch.zeros(
        size=result_shape, dtype=values.dtype, device=values.device
    )

    # construct 2d grid of frequencies and find 2d indices for elements in each bin
    grid = fftfreq_grid(
        image_shape=image_shape[-2:],
        rfft=rfft,
        fftshift=fftshifted,
        norm=True,
        device=values.device,
    )
    split_values = _frequency_bin_split_values(n=values.shape[-1], device=values.device)
    shell_idx = _find_shell_indices_2d(values=grid, split_values=split_values)[:-1]

    # insert data into each shell
    for idx, shell in enumerate(shell_idx):
        idx_h, idx_w = einops.rearrange(shell, "b idx -> idx b")
        average_2d[..., idx_h, idx_w] = values[..., [idx]]

    # fill outside the nyquist circle with the value from the nyquist bin
    average_2d[..., grid > 0.5] = values[..., [-1]]
    return average_2d


def _1d_to_rotational_average_3d_dft(
    values: torch.Tensor,
    image_shape: tuple[int, ...],
    rfft: bool = False,
    fftshifted: bool = True,
) -> torch.Tensor:
    """
    Convert a 1D array of values to a 3D array of values in frequency bins.

    Parameters
    ----------
    values : torch.Tensor
        1D tensor of values to convert
    image_shape : tuple[int, ...]
        Shape of the input image
    rfft : bool
        Whether the input is from an rfft (True) or full fft (False)
    fftshifted : bool
        Whether the input is fftshifted

    Returns
    -------
    torch.Tensor
        3D tensor of values in frequency bins
    """
    # construct output tensor
    d, h, w = image_shape[-3:]
    d, h, w = rfft_shape((d, h, w)) if rfft is True else (d, h, w)
    result_shape = (*image_shape[:-3], d, h, w)
    average_3d = torch.zeros(
        size=result_shape, dtype=values.dtype, device=values.device
    )

    # construct 3d grid of frequencies and find 3d indices for elements in each bin
    grid = fftfreq_grid(
        image_shape=image_shape[-3:],
        rfft=rfft,
        fftshift=fftshifted,
        norm=True,
        device=values.device,
    )
    split_values = _frequency_bin_split_values(n=values.shape[-1], device=values.device)
    shell_idx = _find_shell_indices_3d(values=grid, split_values=split_values)[:-1]

    # insert data into each shell
    for idx, shell in enumerate(shell_idx):
        idx_d, idx_h, idx_w = einops.rearrange(shell, "b idx -> idx b")
        average_3d[..., idx_d, idx_h, idx_w] = values[..., [idx]]

    # fill outside the nyquist circle with the value from the nyquist bin
    average_3d[..., grid > 0.5] = values[..., [-1]]
    return average_3d


def _frequency_bin_centers(n: int, device: torch.device | None = None) -> torch.Tensor:
    """
    Centers of DFT sample frequency bins.

    Parameters
    ----------
    n : int
        Number of frequency bins
    device : torch.device
        Device to place the tensor on

    Returns
    -------
    torch.Tensor
        Centers of the frequency bins
    """
    return torch.linspace(0, 0.5, steps=n, device=device)


def _frequency_bin_split_values(
    n: int,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Values at the borders of DFT sample frequency bins.

    Parameters
    ----------
    n : int
        Number of frequency bins
    device : torch.device
        Device to place the tensor on

    Returns
    -------
    torch.Tensor
        Borders of the frequency bins
    """
    bin_centers = _frequency_bin_centers(n, device=device)
    df = torch.atleast_1d(bin_centers[1])
    bin_centers = torch.concatenate([bin_centers, 0.5 + df], dim=0)  # (b+1, )
    adjacent_bins = bin_centers.unfold(dimension=0, size=2, step=1)  # (b, 2)
    return einops.reduce(adjacent_bins, "b high_low -> b", reduction="mean")


def rfft_shape(
    input_shape: Sequence[int],
) -> tuple[int, ...]:
    """
    Get the output shape of an rfft on an input with input_shape.

    Parameters
    ----------
    input_shape : Sequence[int]
        The shape of the input tensor

    Returns
    -------
    Tuple[int]
        The shape of the output tensor
    """
    rfft_shape = list(input_shape)
    rfft_shape[-1] = int((rfft_shape[-1] / 2) + 1)
    return tuple(rfft_shape)


def bin_or_interpolate_to_output_size(
    values: torch.Tensor,
    output_shape: tuple[int, ...],
) -> torch.Tensor:
    """
    Interpolate 1D frequency values to match the desired output size.

    Parameters
    ----------
    values: torch.Tensor
        1D tensor containing frequency values
    output_shape: tuple[int, ...]
        Desired output shape. The longest dimension will be used as target size.

    Returns
    -------
    torch.Tensor
        Interpolated values matching the target size
    """
    # Get target size (use the longest dimension from output_shape)
    target_size = max(output_shape)

    # If input is batched, handle accordingly
    if values.dim() > 1:
        orig_shape = values.shape
        values = values.view(-1, values.shape[-1])

        # Interpolate each batch
        result = F.interpolate(
            values.unsqueeze(1),  # Add channel dim for interpolate
            size=target_size,
            mode="linear",
            align_corners=True,
        ).squeeze(1)  # Remove channel dim

        # Restore batch dimensions
        return result.view(*orig_shape[:-1], target_size)
    else:
        # Interpolate
        return F.interpolate(
            values.view(1, 1, -1),  # Add batch and channel dims
            size=target_size,
            mode="linear",
            align_corners=True,
        ).view(target_size)  # Remove batch and channel dims
